from typing import Optional
import os
import platform

import click
import logging

from libertem.executor.cli import preload_help

log_values = "Allowed values are 'critical', 'error', 'warning', 'info', 'debug'."


def get_token(token_path):
    token = None
    if token_path is not None:
        with open(token_path) as f:
            token = f.read().strip()
        if len(token) == 0:
            raise click.UsageError(
                f'{token_path} is empty! Refusing to start with insecure configuration.'
            )
    return token


@click.command()
@click.option('-h', '--host', help='host on which the server should listen on',
              default="localhost", type=str, show_default=True)
@click.option('-p', '--port', help='port on which the server should listen on, [default: 9000]',
              default=None, type=int)
@click.option('-d', '--local-directory', help='local directory to manage dask-worker-space files',
              default='dask-worker-space', type=str)
@click.option('-b/-n', '--browser/--no-browser',
              help='enable/disable opening the browser', default='True')
@click.option('-c', '--cpus', help='number of cpu worker processes to use,[default: select in GUI]',
              default=None, type=int)
@click.option('-g', '--gpus', help='number of gpu worker processes to use,[default: select in GUI]',
              default=None, type=int)
@click.option('-o', '--open-ds', help='dataset to preload via URL action',
              default=None, type=str)
@click.option('-l', '--log-level',
              help=f"set logging level. Default is 'info'. {log_values}",
              default='INFO')
@click.option('-t', '--token-path',
              help="path to a file containing a token for authenticating API requests",
              type=click.Path(exists=True))
@click.option('--preload', help=preload_help, default=(), type=str, multiple=True)
@click.option('--insecure',
              help=(
                  "Allow connections from non-localhost without token authorization. "
                  "Applies only when combined with --host <address> "
                  "(use `--insecure -h 0.0.0.0` to accept any connection)"
              ),
              default=False, is_flag=True)
@click.option('--snooze-timeout', type=float,
              help=(
                'Free resources after periods of no activity, in minutes. '
                'Depending on your system, re-starting these resources might '
                'take some time, so typical values are 10 to 30 minutes.'
              ),
              default=None)
def main(port, local_directory, browser, cpus, gpus, open_ds, log_level,
         insecure, host="localhost", token_path=None, preload: tuple[str, ...] = (),
         snooze_timeout: Optional[float] = None):
    # Mitigation for https://stackoverflow.com/questions/71283820/
    #   directory-parameter-on-windows-has-trailing-backslash-replaced-with-double-quote
    if (open_ds and platform.system() == 'Windows' and open_ds[-1] == '"'
            and not os.path.exists(open_ds) and os.path.exists(open_ds[:-1])):
        open_ds = open_ds[:-1]
    # Mitigation for https://github.com/python/cpython/issues/88141
    if platform.system() == 'Windows':
        # Replace the mimetype for .js, as there can be potentially broken
        # entries in the windows registry:
        import mimetypes
        mimetypes.add_type('application/javascript', '.js', strict=True)
    is_custom_port = port is not None
    if port is None:
        port = 9000

    executor_spec = None
    if cpus is not None or gpus is not None:
        executor_spec = dict(
            cpus=cpus if cpus is not None else 0,
            cudas=gpus if gpus is not None else 0,
        )

    from libertem.common.threading import set_num_threads_env
    from libertem.common.tracing import maybe_setup_tracing
    token = get_token(token_path)
    from .server import is_localhost
    if token is None and not is_localhost(host) and not insecure:
        raise click.UsageError(
            f'listening on non-localhost {host}:{port} currently requires token authentication '
            f'or --insecure'
        )
    maybe_setup_tracing(service_name="libertem-server")
    with set_num_threads_env(1):
        from libertem.cli_tweaks import console_tweaks
        from .server import run
        console_tweaks()
        numeric_level = getattr(logging, log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise click.UsageError(f'Invalid log level: {log_level}.\n{log_values}')
        if snooze_timeout is not None:
            snooze_timeout *= 60.0
        run(
            host, port, browser, local_directory, numeric_level,
            token, preload, is_custom_port, executor_spec, open_ds,
            snooze_timeout,
        )


# to enable calling this without using the entry point script,
# for example using `python -m libertem.web.cli`:
if __name__ == "__main__":
    main()



import numpy as np
import pytest

from libertem.udf import UDF
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


class EchoUDF(UDF):
    def get_result_buffers(self):
        print("get_result_buffers", self.params.aux.shape)
        return {
            'echo': self.buffer(
                kind="nav", dtype="float32", extra_shape=(2,)
            ),
            'echo_preprocess': self.buffer(
                kind="nav", dtype="float32", extra_shape=(2,)
            ),
            'echo_postprocess': self.buffer(
                kind="nav", dtype="float32", extra_shape=(2,)
            ),
            'weighted': self.buffer(
                kind="nav", dtype="float32",
            )
        }

    def preprocess(self):
        print("preprocess", self.params.aux.shape)
        self.results.echo_preprocess[:] = self.params.aux

    def process_frame(self, frame):
        print("process_frame", self.params.aux.shape)
        self.results.echo[:] = self.params.aux
        self.results.weighted[:] = np.sum(frame) * self.params.aux[0]

    def postprocess(self):
        print("postprocess", self.params.aux.shape)
        self.results.echo_postprocess[:] = self.params.aux


class EchoTiledUDF(UDF):
    def get_result_buffers(self):
        print("get_result_buffers", self.params.aux.shape)
        return {
            'echo': self.buffer(
                kind="nav", dtype="float32", extra_shape=(2,)
            ),
            'weighted': self.buffer(
                kind="nav", dtype="float32",
            ),
        }

    def process_tile(self, tile):
        print(
            "process_tile",
            self.params.aux.shape,
            tile.shape,
            self.results.echo.shape,
            self.results.weighted.shape
        )
        self.results.echo[:] = self.params.aux
        assert len(self.params.aux2.shape) == 1
        assert len(self.results.weighted.shape) == 1
        w = np.sum(tile, axis=(-1, -2)) * self.params.aux[..., 0]
        self.results.weighted[:] += w


def test_aux_1(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    aux_data = EchoUDF.aux_data(
        kind="nav",
        data=_mk_random(size=(16, 16, 2), dtype="float32"),
        dtype="float32", extra_shape=(2,)
    )
    dataset = lt_ctx.load(
        "memory", data=data, tileshape=(7, 16, 16),
        num_partitions=2, sig_dims=2
    )

    echo_udf = EchoUDF(aux=aux_data, other_stuff=object())
    res = lt_ctx.run_udf(dataset=dataset, udf=echo_udf)
    assert 'echo_preprocess' in res
    print(data.shape, res['echo_preprocess'].data.shape)
    assert np.allclose(res['echo_preprocess'].raw_data, aux_data.raw_data)
    assert 'echo' in res
    print(data.shape, res['echo'].data.shape)
    assert np.allclose(res['echo'].raw_data, aux_data.raw_data)
    assert 'echo_postprocess' in res
    print(data.shape, res['echo_postprocess'].data.shape)
    assert np.allclose(res['echo_postprocess'].raw_data, aux_data.raw_data)


def test_aux_roi_dummy(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    aux_input = _mk_random(size=(16, 16, 2), dtype="float32")
    aux_data = EchoUDF.aux_data(
        kind="nav",
        data=aux_input,
        dtype="float32", extra_shape=(2,)
    )
    dataset = lt_ctx.load(
        "memory", data=data, tileshape=(7, 16, 16),
        num_partitions=2, sig_dims=2
    )

    roi = np.ones(dataset.shape.nav, dtype="bool")

    echo_udf = EchoUDF(aux=aux_data)
    res = lt_ctx.run_udf(dataset=dataset, udf=echo_udf, roi=roi)
    assert 'echo_preprocess' in res
    print(data.shape, res['echo_preprocess'].data.shape)
    assert np.allclose(res['echo_preprocess'].data[roi], aux_input[roi])
    assert 'echo' in res
    print(data.shape, res['echo'].data.shape)
    assert np.allclose(res['echo'].data[roi], aux_input[roi])
    assert 'echo_postprocess' in res
    print(data.shape, res['echo_postprocess'].data.shape)
    assert np.allclose(res['echo_postprocess'].data[roi], aux_input[roi])


def test_aux_roi(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    aux_input = _mk_random(size=(16, 16, 2), dtype="float32")
    aux_data = EchoUDF.aux_data(
        kind="nav",
        data=aux_input,
        dtype="float32", extra_shape=(2,)
    )
    dataset = lt_ctx.load(
        "memory", data=data, tileshape=(7, 16, 16),
        num_partitions=2, sig_dims=2
    )

    roi = _mk_random(size=dataset.shape.nav, dtype="bool")

    echo_udf = EchoUDF(aux=aux_data)
    res = lt_ctx.run_udf(dataset=dataset, udf=echo_udf, roi=roi)
    assert 'echo_preprocess' in res
    print(data.shape, res['echo_preprocess'].data.shape)
    assert np.allclose(res['echo_preprocess'].data[roi], aux_input[roi])
    assert 'echo' in res
    print(data.shape, res['echo'].data.shape)
    assert np.allclose(res['echo'].data[roi], aux_input[roi])
    assert 'echo_postprocess' in res
    print(data.shape, res['echo_postprocess'].data.shape)
    assert np.allclose(res['echo_postprocess'].data[roi], aux_input[roi])


def test_aux_2(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    aux_data = EchoUDF.aux_data(
        kind="nav", dtype="float32", extra_shape=(2,),
        data=_mk_random(size=(16, 16, 2), dtype="float32"),
    )
    dataset = MemoryDataSet(data=data, tileshape=(7, 16, 16),
                            num_partitions=2, sig_dims=2)

    echo_udf = EchoUDF(aux=aux_data)
    res = lt_ctx.run_udf(dataset=dataset, udf=echo_udf)
    assert 'weighted' in res
    print(data.shape, res['weighted'].data.shape)
    assert np.allclose(
        res['weighted'].raw_data,
        np.sum(data, axis=(2, 3)).reshape(-1) * aux_data.raw_data[..., 0]
    )


@pytest.mark.parametrize(
    "num_partitions", (2, 4)
)
def test_aux_small(num_partitions, lt_ctx):
    data = _mk_random(size=(2, 2, 16, 16), dtype="float32")
    aux_data = EchoUDF.aux_data(
        kind="nav", dtype="float32", extra_shape=(2, ),
        data=_mk_random(size=(2, 2, 2), dtype="float32"),
    )
    dataset = MemoryDataSet(data=data, tileshape=(7, 16, 16),
                            num_partitions=num_partitions, sig_dims=2)

    echo_udf = EchoUDF(aux=aux_data)
    res = lt_ctx.run_udf(dataset=dataset, udf=echo_udf)
    assert 'weighted' in res
    print(data.shape, res['weighted'].data.shape)
    assert np.allclose(
        res['weighted'].raw_data,
        np.sum(data, axis=(2, 3)).reshape(-1) * aux_data.raw_data[..., 0]
    )


def test_aux_constructor(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    aux_data = EchoUDF.aux_data(
        data=_mk_random(size=(16, 16, 2), dtype="float32"),
        kind="nav", dtype="float32", extra_shape=(2,)
    )
    dataset = MemoryDataSet(data=data, tileshape=(7, 16, 16),
                            num_partitions=2, sig_dims=2)

    echo_udf = EchoUDF(aux=aux_data)
    res = lt_ctx.run_udf(dataset=dataset, udf=echo_udf)
    assert 'weighted' in res
    print(data.shape, res['weighted'].data.shape)
    assert np.allclose(
        res['weighted'].raw_data,
        np.sum(data, axis=(2, 3)).reshape(-1) * aux_data.raw_data[..., 0]
    )


@pytest.mark.parametrize(
    'tileshape', [(2, 16, 16), (1, 16, 16), (1, 9, 16)]
)
def test_aux_tiled(lt_ctx, tileshape):
    data = _mk_random(size=(1, 5, 16, 16), dtype="float32")
    aux_data = EchoTiledUDF.aux_data(
        data=_mk_random(size=(1, 5, 2), dtype="float32"),
        kind="nav", dtype="float32", extra_shape=(2,)
    )
    aux_data_2 = EchoTiledUDF.aux_data(
        data=_mk_random(size=(1, 5), dtype="float32"),
        kind="nav", dtype="float32"
    )
    dataset = MemoryDataSet(data=data, tileshape=tileshape,
                            num_partitions=2, sig_dims=2)

    echo_udf = EchoTiledUDF(aux=aux_data, aux2=aux_data_2)
    res = lt_ctx.run_udf(dataset=dataset, udf=echo_udf)
    assert 'weighted' in res
    print(data.shape, res['weighted'].data.shape)
    assert np.allclose(
        res['weighted'].raw_data,
        np.sum(data, axis=(2, 3)).reshape(-1) * aux_data.raw_data[..., 0]
    )



import time
import concurrent

import pytest
import numpy as np

from libertem.common.async_utils import async_generator_eager, async_generator
from libertem.common.executor import SimpleMPWorkerQueue, SimpleWorkerQueue
from libertem.common.buffers import reshaped_view


@pytest.fixture(scope='module')
def thread_pool():
    pool = concurrent.futures.ThreadPoolExecutor(1)
    # warm up:
    pool.submit(lambda: 1).result()
    return pool


@pytest.mark.asyncio
async def test_async_generator_eager(thread_pool):
    results = []

    def sync_generator():
        for i in range(20):
            results.append('result')
            yield 'result'

    gen = sync_generator()
    async_gen = async_generator_eager(gen, pool=thread_pool)

    await async_gen.__anext__()
    time.sleep(0.1)

    # the generator has run to completion already:
    assert len(results) == 20

    async for value in async_gen:
        assert value == 'result'

    results = []

    gen = sync_generator()
    async_gen = async_generator(gen, pool=thread_pool)

    await async_gen.__anext__()
    time.sleep(0.1)

    # the generator has only run a single iteration:
    assert len(results) == 1

    async for value in async_gen:
        assert value == 'result'


@pytest.mark.asyncio
async def test_async_generator_err_handling(thread_pool):
    def sync_generator():
        for i in range(2):
            yield 'result'
        raise RuntimeError("something unexpected happened")

    gen = sync_generator()
    async_gen = async_generator_eager(gen, pool=thread_pool)

    with pytest.raises(RuntimeError) as e:
        async for value in async_gen:
            assert value == 'result'

    assert e.match("something unexpected happened")


@pytest.mark.parametrize(
    'queue', [SimpleMPWorkerQueue(), SimpleWorkerQueue()]
)
def test_worker_queues(queue):
    payload_data = np.random.random((23, 42))
    header = {
        'asdf': lambda x: x,
        'abc': 23,
        'shape': payload_data.shape,
        'dtype': payload_data.dtype,
    }

    def check_received(received_header, decoded_payload):
        assert received_header['asdf']('abc') == 'abc'
        for key in ('abc', 'shape', 'dtype'):
            assert header[key] == received_header[key]
        assert np.all(payload_data == decoded_payload)

    queue.put(header, payload_data)
    with queue.get() as res:
        check_received(*res)

    with queue.put_nocopy(header, size=payload_data.nbytes) as payload:
        payload[:] = reshaped_view(payload_data, (23*42, )).view(np.uint8)

    with queue.get() as res:
        received_header, received_payload = res
        decoded_payload = reshaped_view(
            np.asarray(received_payload).view(received_header['dtype']),
            received_header['shape']
        )
        check_received(received_header, decoded_payload)



import re
import os
import codecs
import subprocess
import distutils
import pathlib
import shutil
from setuptools.command.sdist import sdist
from setuptools import setup, find_packages


class BuildClientCommand(distutils.cmd.Command):
    description = 'build the js client'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cwd = os.path.dirname(__file__)
        cwd_client = os.path.join(cwd, 'client')
        self.announce(
            "building js client",
            level=distutils.log.INFO
        )
        npm = shutil.which('npm')
        for command in [[npm, 'install'],
                        [npm, 'run-script', 'build']]:
            self.announce(' '.join(command), distutils.log.INFO)
            subprocess.check_call(command, cwd=cwd_client)
        self.run_command('copy_client')


class CopyClientCommand(distutils.cmd.Command):
    description = 'copy the js client'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cwd = pathlib.Path(__file__).absolute().parent
        cwd_client = cwd / 'client'
        client = cwd / 'src' / 'libertem' / 'web' / 'client'

        self.announce(
            "preparing output directory: %s" % client,
            level=distutils.log.INFO
        )
        if client.exists():
            shutil.rmtree(client)

        build = cwd_client / "dist"
        self.announce(
            f"copying client: {build} -> {client}",
            level=distutils.log.INFO
        )
        shutil.copytree(build, client)


class BakedRevisionBuilderSdist(sdist):
    def make_release_tree(self, base_dir, files):
        if not self.dry_run:
            write_baked_revision(base_dir)
        super().make_release_tree(base_dir, files)


def mkpath(path):
    if not os.path.exists(path):
        os.makedirs(path)


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    here = os.path.abspath(os.path.dirname(__file__))
    return codecs.open(os.path.join(here, *parts), 'r').read()


def remove_rst_roles(txt):
    return re.sub(':(cite|doc):`[^`]+` ?', '', txt)


def get_git_rev():
    # NOTE: this is a copy from src/libertem/versioning.py
    # this is because it is not guaranteed that we can import our own packages
    # from setup.py AFAIK
    try:
        new_cwd = os.path.abspath(os.path.dirname(__file__))
        rev_raw = subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=new_cwd)
        return rev_raw.decode("utf8").strip()
    except Exception:
        return "unknown"


def write_baked_revision(base_dir):
    dest_dir = os.path.join(base_dir, 'src', 'libertem')
    baked_dest = os.path.join(dest_dir, '_baked_revision.py')
    mkpath(dest_dir)

    with open(baked_dest, "w") as f:
        f.write(r'revision = "%s"' % get_git_rev())


def find_version(*file_paths):
    """
    "stolen" from pip's setup.py
    """
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="libertem",
    version=find_version("src", "libertem", "__version__.py"),
    license='GPL v3',
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.9.3',
    install_requires=[
        "numpy",
        "scipy>=1.4.1",
        "sparse",
        "distributed>=2.19.0",
        "dask!=2023.6.1",  # work around broken ipython exception handler
        "click",
        "tornado>=5",
        "matplotlib",
        "pillow",
        "h5py",
        "psutil",
        # https://github.com/LiberTEM/LiberTEM/issues/763
        # https://github.com/LiberTEM/LiberTEM/issues/783
        "numba>=0.51",
        "ncempy>=1.10",
        'pywin32!=226;platform_system=="Windows"',
        # FIXME pull request #259
        # https://github.com/LiberTEM/LiberTEM/pull/259#discussion_r251877431
        'scikit-image',
        'cloudpickle',
        'jsonschema',
        'scikit-learn',
        'tqdm',
        'threadpoolctl>=3.0',
        'nbformat',
        'nbconvert',
        'autopep8',
        'empyre>=0.3.0',
        'defusedxml',
        'typing-extensions',  # backwards-compatibility for newer typing constructs
        'opentelemetry-api',
        'tblib',
        'tomli',
        'sparseconverter>=0.3.3',
        'numexpr!=2.8.6',  # work around broken sanitization, remove this line when fixed
        'jupyter_ui_poll',  # Make live matplotlib plots work with the widget backend
        'ipympl',  # Support the widget backend
    ],
    extras_require={
        # NumPy interfacing issue on Win 11, Python 3.10
        # Likely caused by other packages constraining NumPy version to <1.23
        # FIXME recheck later again
        'torch': 'torch<1.12',
        # Typing issue in hdbscan, maybe
        # https://github.com/scikit-learn-contrib/hdbscan/issues/600
        # FIXME check if error in tests\test_fullmatch.py persists in a later
        # hdbscan release
        'hdbscan': [
            "hdbscan<=0.8.30;( python_version=='3.11' and platform_system=='Windows')",
            "hdbscan;( python_version!='3.11' or platform_system!='Windows')",
        ],
        'cupy': 'cupy',
        'bqplot': ['bqplot', 'bqplot-image-gl', 'ipython'],
        'hdf5plugin': 'hdf5plugin',
        'tracing': [
            'opentelemetry-distro',
            'opentelemetry-exporter-otlp',
        ],
    },
    package_dir={"": "src"},
    packages=find_packages(where='src'),
    entry_points={
        'console_scripts': [
            'libertem-server=libertem.web.cli:main',
            'libertem-worker=libertem.executor.cli:main',
        ]
    },
    cmdclass={
        'build_client': BuildClientCommand,
        'copy_client': CopyClientCommand,
        'sdist': BakedRevisionBuilderSdist,
    },
    keywords="electron microscopy",
    description="Open pixelated STEM framework",
    long_description=remove_rst_roles(read("README.rst")),
    long_description_content_type="text/x-rst",
    url="https://libertem.github.io/LiberTEM/",
    author_email="libertem-dev@googlegroups.com",
    author="the LiberTEM team",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Environment :: Web Environment',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: JavaScript',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
)



import os
import mmap
import contextlib

import numpy as np
from numba.typed import List
import numba
from sparseconverter import CUDA, NUMPY, ArrayBackend

from libertem.common.math import prod
from libertem.io.dataset.base.backend import IOBackend, IOBackendImpl
from libertem.io.dataset.base.fileset import FileSet
from libertem.common import Shape, Slice
from libertem.common.buffers import BufferPool
from libertem.common.numba import cached_njit
from .tiling import DataTile
from .decode import DtypeConversionDecoder


_r_n_d_cache = {}


def _make_mmap_reader_and_decoder(decode):
    """
    decode: from inp, in bytes, possibly interpreted as native_dtype, to out_decoded.dtype
    """
    @cached_njit(boundscheck=False, cache=True, nogil=True)
    def _mmap_tilereader_w_copy(outer_idx, mmaps, sig_dims, tile_read_ranges,
                           out_decoded, native_dtype, do_zero,
                           origin, shape, ds_shape):
        """
        Read and decode a single tile
        """
        if do_zero:
            out_decoded[:] = 0
        for rr_idx in range(tile_read_ranges.shape[0]):
            rr = tile_read_ranges[rr_idx]
            if rr[1] == rr[2] == 0:
                break
            memmap = mmaps[rr[0]]
            decode(
                inp=memmap[rr[1]:rr[2]],
                out=out_decoded,
                idx=rr_idx,
                native_dtype=native_dtype,
                rr=rr,
                origin=origin,
                shape=shape,
                ds_shape=ds_shape,
            )
        return out_decoded
    return _mmap_tilereader_w_copy


@numba.njit(nogil=True)
def _get_prefetch_ranges(num_files, tile_ranges):
    res = np.zeros((num_files, 3), dtype=np.int64)
    for rr_idx in range(tile_ranges.shape[0]):
        rr = tile_ranges[rr_idx]
        if res[rr[0], 0] == 0:
            res[rr[0], 0] = rr[1]  # if minimum is 0, pick first value
        res[rr[0], 0] = min(res[rr[0], 0], rr[1])
        res[rr[0], 1] = max(res[rr[0], 1], rr[2])
        res[rr[0], 2] = rr[0]
    return res


class MMapBackend(IOBackend, id_="mmap"):
    """
    I/O backend using memory mapped files. Used by default on non-Windows
    systems.

    Parameters
    ----------
    enable_readahead_hints : bool
        Linux only. Try to influence readahead behavior (experimental).
    """
    def __init__(self, enable_readahead_hints=False):
        self._enable_readahead = enable_readahead_hints

    def get_impl(self):
        return MMapBackendImpl(self._enable_readahead)

    @classmethod
    def from_json(cls, msg):
        """
        Construct an instance from the already-decoded `msg`.
        """
        raise NotImplementedError("TODO! implement me!")


class MMapFileBase:
    def __init__(self, path, desc):
        raise NotImplementedError()

    def open(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    @property
    def mmap(self):
        raise NotImplementedError()

    @property
    def array(self):
        raise NotImplementedError()

    @property
    def handle(self):
        raise NotImplementedError()


class MMapFile(MMapFileBase):
    def __init__(self, path, desc):
        self.path = path
        self.desc = desc
        self._handle = None
        self._mmap = None
        self._arr = None

    def open(self):
        # FIXME: maybe DO switch off buffering? because it means more copies, right?
        # NOTE: for `readinto` to work, we must not switch off buffering here!
        # otherwise, `readinto` may return partial results, which can be hard to handle
        self._handle = open(self.path, "rb")
        self._mmap = mmap.mmap(
            fileno=self._handle.fileno(),
            length=0,
            # can't use offset for cutting off file header, as it needs to be
            # aligned to page size...
            offset=0,
            access=mmap.ACCESS_READ,
        )
        slicing = self.desc.get_offsets_sizes(self._mmap.size())
        self._arr = self.desc.get_array_from_memview(
            memoryview(self._mmap),
            slicing
        )
        return self

    def close(self):
        del self._arr
        del self._mmap
        self._handle.close()
        self._handle = None

    @property
    def mmap(self):
        if self._mmap is None:
            raise RuntimeError("trying to access a mmap for a closed file")
        return self._mmap

    @property
    def array(self):
        if self._arr is None:
            raise RuntimeError("trying to access array for a closed file")
        return self._arr

    @property
    def handle(self):
        if self._handle is None:
            raise RuntimeError("trying to access file handle of closed file")
        return self._handle


class MMapBackendImpl(IOBackendImpl):
    FILE_CLS: type = MMapFile

    def __init__(self, enable_readahead_hints=False):
        super().__init__()
        self._enable_readahead = enable_readahead_hints
        self._buffer_pool = BufferPool()

    @contextlib.contextmanager
    def open_files(self, fileset: FileSet):
        mmap_files = [
            self.FILE_CLS(path=f.path, desc=f).open()
            for f in fileset
        ]
        yield mmap_files
        for f in mmap_files:
            f.close()

    def _get_tiles_straight(
        self, tiling_scheme, open_files: list[MMapFile], read_ranges,
        sync_offset=0
    ):
        """
        Read straight from the file system cache, via memory mapping, without
        any decoding step.

        This method makes a few assumptions:
         - no corrections are needed
         - tiles don't span multiple files
         - the `LocalFile` has already cut away headers and footers
           (both per-file and per-frame) from the mmap

        Parameters
        ----------

        tiling_scheme : TilingScheme

        open_files : List[MMapFile]
            The list of files must correspond to the indices used in the `read_ranges`.
            Usually, that means it is limited to the files that are part of the
            current partition.

        read_ranges : Tuple[np.ndarray, np.ndarray, np.ndarray]
            As returned by `get_read_ranges`

        sync_offset : int
        """

        ds_sig_shape = tuple(tiling_scheme.dataset_shape.sig)
        sig_dims = tiling_scheme.shape.sig.dims
        slices, ranges, scheme_indices = read_ranges
        for idx in range(slices.shape[0]):
            origin, shape = slices[idx]
            tile_ranges = ranges[idx]
            scheme_idx = scheme_indices[idx]

            # NOTE: for straight mmap, read_ranges must not contain tiles
            # that span multiple files. This is ensured in IOBackend.need_copy
            # (that is, we force copying if files are smaller than tiles)
            file_idx = tile_ranges[0][0]
            fh = open_files[file_idx]
            arr = fh.array.reshape((fh.desc.num_frames,) + ds_sig_shape)
            tile_slice = Slice(
                origin=origin,
                shape=Shape(shape, sig_dims=sig_dims)
            )
            # sync_offset is either zero or positive
            # in case of negative sync_offset, _get_tiles_w_copy is used
            data_slice = (
                slice(
                    origin[0] - fh.desc.start_idx + sync_offset,
                    origin[0] - fh.desc.start_idx + shape[0] + sync_offset
                ),
            ) + tuple(
                slice(o, (o + s))
                for (o, s) in zip(origin[1:], shape[1:])
            )
            data = arr[data_slice]
            yield DataTile(
                data,
                tile_slice=tile_slice,
                scheme_idx=scheme_idx,
            )

    def get_read_and_decode(self, decode):
        key = (decode, "mmap")
        if key in _r_n_d_cache:
            return _r_n_d_cache[key]
        r_n_d = _make_mmap_reader_and_decoder(decode=decode)
        _r_n_d_cache[key] = r_n_d
        return r_n_d

    def _get_tiles_w_copy(
        self, tiling_scheme, open_files, read_ranges, read_dtype, native_dtype,
        decoder=None, corrections=None,
    ):
        if decoder is None:
            decoder = DtypeConversionDecoder()
        decode = decoder.get_decode(
            native_dtype=np.dtype(native_dtype),
            read_dtype=np.dtype(read_dtype),
        )
        r_n_d = self._r_n_d = self.get_read_and_decode(decode)

        native_dtype = decoder.get_native_dtype(native_dtype, read_dtype)
        mmaps = List()
        for fh in open_files:
            mmaps.append(np.frombuffer(fh.mmap, dtype=np.uint8))

        sig_dims = tiling_scheme.shape.sig.dims
        ds_shape = np.array(tiling_scheme.dataset_shape)

        largest_slice = sorted((
            (prod(s_.shape), s_)
            for _, s_ in tiling_scheme.slices
        ), key=lambda x: x[0], reverse=True)[0][1]

        buf_shape = (tiling_scheme.depth,) + tuple(largest_slice.shape)
        need_clear = decoder.do_clear()

        with self._buffer_pool.empty(buf_shape, dtype=read_dtype) as out_decoded:
            out_decoded = out_decoded.reshape((-1,))
            slices = read_ranges[0]
            # Use NumPy prod for multidimensional array and axis parameter
            shape_prods = np.prod(slices[..., 1, :], axis=1, dtype=np.int64)
            ranges = read_ranges[1]
            scheme_indices = read_ranges[2]
            for idx in range(slices.shape[0]):
                origin = slices[idx, 0]
                shape = slices[idx, 1]
                tile_slice = Slice(
                    origin=origin,
                    shape=Shape(shape, sig_dims=sig_dims)
                )
                tile_ranges = ranges[idx]
                scheme_idx = scheme_indices[idx]
                # if idx < slices.shape[0] - 1:
                #     self._prefetch_for_tile(fileset, ranges[idx + 1])
                #     pass
                out_cut = out_decoded[:shape_prods[idx]].reshape((shape[0], -1))
                data = r_n_d(
                    idx,
                    mmaps, sig_dims, tile_ranges,
                    out_cut, native_dtype, do_zero=need_clear,
                    origin=origin, shape=shape, ds_shape=ds_shape,
                )
                data = data.reshape(shape)
                self.preprocess(data, tile_slice, corrections)
                yield DataTile(
                    data,
                    tile_slice=tile_slice,
                    scheme_idx=scheme_idx,
                )

    def get_tiles(
        self, tiling_scheme, fileset, read_ranges, roi, native_dtype, read_dtype, decoder,
        sync_offset, corrections, array_backend: ArrayBackend,
    ):
        # TODO: how would compression work?
        # TODO: sparse input data? COO format? fill rate? → own pipeline! → later!
        # strategy: assume low (20%?) fill rate, read whole partition and apply ROI in-memory
        #           partitioning when opening the dataset, or by having one file per partition
        assert array_backend in (NUMPY, CUDA)
        with self.open_files(fileset) as open_files:
            if self._enable_readahead:
                self._set_readahead_hints(roi, open_files)
            if not self.need_copy(
                decoder=decoder,
                tiling_scheme=tiling_scheme,
                fileset=fileset,
                roi=roi,
                native_dtype=native_dtype,
                read_dtype=read_dtype,
                sync_offset=sync_offset,
                corrections=corrections,
            ):
                yield from self._get_tiles_straight(
                    tiling_scheme, open_files, read_ranges, sync_offset,
                )
            else:
                yield from self._get_tiles_w_copy(
                    tiling_scheme=tiling_scheme,
                    open_files=open_files,
                    read_ranges=read_ranges,
                    read_dtype=read_dtype,
                    native_dtype=native_dtype,
                    decoder=decoder,
                    corrections=corrections,
                )

    def _prefetch_for_tile(self, fileset, tile_ranges):
        prefr = _get_prefetch_ranges(len(fileset), tile_ranges)
        prefr = prefr[~np.all(prefr == 0, axis=1)]
        for mi, ma, fidx in prefr:
            f = fileset[fidx]
            os.posix_fadvise(
                f.handle.fileno(),
                mi,
                ma - mi,
                os.POSIX_FADV_WILLNEED
            )

    def _set_readahead_hints(self, roi, open_files):
        if not hasattr(os, 'posix_fadvise'):
            return
        if any([f.handle.fileno() is None
                for f in open_files]):
            return
        for f in open_files:
            os.posix_fadvise(
                f.handle.fileno(),
                0,
                0,
                os.POSIX_FADV_WILLNEED
            )



import io
import os

import numpy as np
import pytest
import nbformat
from temp_utils import _get_hdf5_params
from libertem.udf.stddev import StdDevUDF
from libertem import masks
from libertem.web.notebook_generator.notebook_generator import notebook_generator
from nbconvert.preprocessors import ExecutePreprocessor


@pytest.mark.slow
def test_sd_default(hdf5_ds_2, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)
    params = {"roi": {}}
    analysis = [{
            "analysisType": 'SD_FRAMES',
            "parameters": params,
            }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    data_path = os.path.join(datadir, 'sd_result.npy')
    results = np.load(data_path)
    udf = StdDevUDF()
    expected = lt_ctx.run_udf(dataset=hdf5_ds_2, udf=udf)
    assert np.allclose(
        results,
        expected['varsum'],
    )


@pytest.mark.slow
def test_sd_roi(hdf5_ds_2, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)

    roi_params = {
        "shape": "rect",
        "x": 1,
        "y": 2,
        "width": 6,
        "height": 6
    }

    analysis = [{
                "analysisType": 'SD_FRAMES',
                "parameters": {
                            "roi": roi_params
                            }
                }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    data_path = os.path.join(datadir, 'sd_result.npy')
    results = np.load(data_path)
    nx, ny = hdf5_ds_2.shape.nav
    roi = masks.rectangular(
                X=roi_params["x"],
                Y=roi_params["y"],
                Width=roi_params["width"],
                Height=roi_params["height"],
                imageSizeX=nx,
                imageSizeY=ny)
    udf = StdDevUDF()
    expected = lt_ctx.run_udf(dataset=hdf5_ds_2, udf=udf, roi=roi)
    assert np.allclose(
        results,
        expected['varsum'],
    )



import warnings

import numpy as np

from .base import Live2DPlot


class GMSLive2DPlot(Live2DPlot):
    """
    Live plot for Gatan Microscopy Suite, Digital Micrograph (experimental).

    This works with Python scripting within GMS

    .. versionadded:: 0.7.0

    Parameters
    ----------
    dataset : DataSet
        The dataset on which the UDF will be run. This allows
        to determine the shape of the plots for initialization.
    udf : UDF
        The UDF instance this plot is associated to. This needs to be
        the same instance that is passed to
        :meth:`~libertem.api.Context.run_udf`.
    roi : numpy.ndarray or None
        Region of interest (ROI) that the UDF will
        be run on. This is necessary for UDFs where the `extra_shape`
        parameter of result buffers is a function of the ROI, such as
        :class:`~libertem.udf.raw.PickUDF`.
    channel : misc
        Indicate the channel to be plotted.

        - :code:`None`: The first plottable (2D) channel of the UDF is plotted.
        - :code:`str`: The UDF result buffer name that should be plotted.
        - :code:`tuple(str, function(ndarray) -> ndarray)`: The UDF result buffer name that
          should be plotted together with a function that extracts a plottable result
        - :code:`function(udf_result, damage) -> (ndarray, damage)`: Function that derives a
          plottable 2D ndarray and damage indicator from the full
          UDF results and the processed nav space. See :ref:`plotting` for more details!
    title : str
        The plot title. By default UDF class name and channel name.
    min_delta : float
        Minimum time span in seconds between updates to reduce overheads for slow plotting.
    udfresult : UDFResults, optional
        UDF result to initialize the plot data and determine plot shape. If None (default),
        this is determined using :meth:`~libertem.udf.base.UDFRunner.dry_run` on the dataset,
        UDF and ROI. This parameter allows re-using buffers to avoid unnecessary dry runs.
    """
    def __init__(
            self, dataset, udf, roi=None, channel=None, title=None,
            min_delta=0.2, udfresult=None,

    ):
        super().__init__(
            dataset=dataset,
            udf=udf,
            roi=roi,
            channel=channel,
            title=title,
            min_delta=min_delta,
            udfresult=udfresult
        )
        # Optional dependency, don't import top-level
        import DigitalMicrograph as DM

        self.image = DM.CreateImage(self.data.copy())
        self.window = None
        self.disp = None
        self.image.SetName(self.title)

    def display(self):
        '''
        Use :meth:`DM.ShowImage` to create an image display.
        '''
        self.window = self.image.ShowImage()
        self.disp = self.image.GetImageDisplay(0)
        self.disp.SetDoAutoSurvey(False)

    def update(self, damage, force=False):
        if self.disp is None:
            assert self.window is None
            warnings.warn(
                "Plot is not displayed, not plotting. "
                "Call display() to display the plot."
            )
            return
        valid_data = self.data[damage]
        if len(valid_data) > 0:
            vmin = np.min(valid_data)
            vmax = np.max(valid_data)
            self.disp.SetContrastLimits(float(vmin), float(vmax))
        buffer = self.image.GetNumArray()
        buffer[:] = self.data
        self.image.UpdateImage()



import multiprocessing as mp
import contextlib
import math
import queue
from typing import TYPE_CHECKING, Any, NamedTuple, Optional
from collections.abc import Generator

import cloudpickle
import numpy as np

from libertem.common.executor import WorkerQueue, WorkerQueueEmpty

if TYPE_CHECKING:
    from multiprocessing import shared_memory


class PoolAllocation(NamedTuple):
    shm_name: str
    handle: int  # internal handle, could be an offset
    full_size: int  # full size of the allocation, in bytes (including padding)
    req_size: int  # requested allocation size

    def resize(self, new_req_size) -> "PoolAllocation":
        assert new_req_size <= self.full_size
        return PoolAllocation(
            shm_name=self.shm_name,
            handle=self.handle,
            full_size=self.full_size,
            req_size=new_req_size,
        )


class PoolShmClient:
    def __init__(self):
        self._shm_cache = {}

    def get(self, allocation: PoolAllocation) -> memoryview:
        from multiprocessing import shared_memory
        offset = allocation.handle
        name = allocation.shm_name
        if name in self._shm_cache:
            shm = self._shm_cache[name]
        else:
            self._shm_cache[name] = shm = shared_memory.SharedMemory(name=name, create=False)
        return shm.buf[offset:offset+allocation.req_size]


class PoolShmAllocator:
    ALIGN_TO = 4096

    def __init__(self, item_size, size_num_items, create=True, name=None):
        """
        Bump plus free list allocator. Can allocate objects of a fixed size.
        Memory above the `_used` offset is free, allocated blocks from the
        `_free` list are free, everything else is in use.

        Allocation and recycling should happen on the same process.
        """
        self._item_size = item_size
        size = item_size * size_num_items
        size = self.ALIGN_TO * math.ceil(size / self.ALIGN_TO)
        self._create = create
        self._name = name
        self._size = size
        self._free = []  # list of byte offsets into `_shm`
        self._used = 0  # byte offset into `_shm`; everything above is free memory
        self._shm = self._open()

    def _open(self):
        from multiprocessing import shared_memory
        shm = shared_memory.SharedMemory(
            create=self._create,
            name=self._name,
            size=self._size
        )
        return shm

    def allocate(self, req_size: int) -> PoolAllocation:
        if req_size > self._item_size:
            raise RuntimeError(f"allocation request for size {req_size} cannot be serviced")
        # 1) check free list, get an item from there
        if len(self._free) > 0:
            offset = self._free.pop()
        # 2) if free list is empty, bump up `_used` and return a "new" block of
        #    memory
        elif self._used + self._item_size < self._size:
            offset = self._used
            self._used += self._item_size
        else:
            raise RuntimeError("pool shm is out of memory")
        return PoolAllocation(
            shm_name=self._shm.name,
            handle=offset,
            full_size=self._item_size,
            req_size=req_size,
        )

    def get(self, allocation: PoolAllocation) -> memoryview:
        offset = allocation.handle
        assert allocation.shm_name == self._shm.name
        return self._shm.buf[offset:offset+allocation.req_size]

    def recycle(self, allocation: PoolAllocation):
        self._free.append(allocation.handle)

    def shutdown(self):
        self._shm.unlink()

    @property
    def name(self):
        return self._shm.name


def drain_queue(q: mp.Queue):
    while True:
        try:
            q.get_nowait()
        except queue.Empty:
            break


class ShmQueue(WorkerQueue):
    def __init__(self):
        mp_ctx = mp.get_context("spawn")
        self.q = mp_ctx.Queue()
        self.release_q = mp_ctx.Queue()
        self._pool_shm_allocator = None
        self._pool_shm_client = None
        self._closed = False

    @contextlib.contextmanager
    def put_nocopy(self, header: Any, size: int) -> Generator[memoryview, None, None]:
        alloc_handle, payload_shm = self._get_buf_for_writing(size)
        yield payload_shm
        self.q.put((cloudpickle.dumps(header), 'bytes', alloc_handle))

    def put(self, header, payload: Optional[memoryview] = None):
        """
        Send the (header, payload) tuple via this channel - copying the
        `payload` to a shared memory segment while sending `header` plainly
        via a queue. The header should be `pickle`able.
        """
        if payload is not None:
            payload_shm: Optional[PoolAllocation] = self._copy_to_shm(payload)
        else:
            payload_shm = None
        self.q.put((cloudpickle.dumps(header), 'bytes', payload_shm))

    def _get_buf_for_writing(self, size: int) -> tuple[PoolAllocation, memoryview]:
        if self._pool_shm_allocator is None:
            # FIXME: config item size, pool size
            self._pool_shm_allocator = PoolShmAllocator(
                item_size=512*512*4*2, size_num_items=24*128
            )
        try:
            alloc_handle: PoolAllocation = self.release_q.get_nowait()
            alloc_handle = alloc_handle.resize(size)
        except queue.Empty:
            alloc_handle = self._pool_shm_allocator.allocate(size)
        payload_shm = self._pool_shm_allocator.get(alloc_handle)
        assert payload_shm.nbytes == size, f"{payload_shm.nbytes} != {size}"
        return alloc_handle, payload_shm

    def _copy_to_shm(self, src_buffer: memoryview) -> PoolAllocation:
        """
        Copy the `buffer` to shared memory and return its name
        """
        size = src_buffer.nbytes
        alloc_handle, payload_shm = self._get_buf_for_writing(size)
        src_arr = np.frombuffer(src_buffer, dtype=np.uint8)
        arr_shm = np.frombuffer(payload_shm, dtype=np.uint8)
        assert arr_shm.size == size, f"{arr_shm.size} != {size}"
        arr_shm[:] = src_arr
        return alloc_handle

    def _get_named_shm(self, name: str) -> "shared_memory.SharedMemory":
        from multiprocessing import shared_memory
        return shared_memory.SharedMemory(name=name, create=False)

    @contextlib.contextmanager
    def get(self, block: bool = True, timeout: Optional[float] = None):
        """
        Receive a message. Memory of the payload will be cleaned up after the
        context manager scope, so don't keep references outside of it!

        Parameters
        ----------
        timeout
            Timeout in seconds,
        """
        if self._pool_shm_client is None:
            self._pool_shm_client = PoolShmClient()
        payload_memview: Optional[memoryview] = None
        payload_handle = None
        try:
            header, typ, payload_handle = self.q.get(block=block, timeout=timeout)
            if payload_handle is not None:
                payload_buf = self._pool_shm_client.get(payload_handle)
                payload_memview = memoryview(payload_buf)
            else:
                payload_buf = None
                payload_memview = None
            if typ == "bytes":
                yield (cloudpickle.loads(header), payload_memview)
        except queue.Empty:
            raise WorkerQueueEmpty()
        finally:
            if payload_memview is not None:
                payload_memview.release()
            if payload_handle is not None:
                self.release_q.put(payload_handle)

    def empty(self):
        return self.q.empty()

    def close(self):
        if not self._closed:
            drain_queue(self.q)
            self.q.close()
            self.q.join_thread()
            drain_queue(self.release_q)
            self.release_q.close()
            self.release_q.join_thread()
            self._closed = True



import numpy as np

from libertem.udf import UDF


class LogsumUDF(UDF):
    """
    Sum up logscaled frames

    In comparison to log-scaling the sum, this highlights regions with slightly higher
    intensity that appear in many frames in relation to very high intensity in a few frames.

    Examples
    --------
    >>> udf = LogsumUDF()
    >>> result = ctx.run_udf(dataset=dataset, udf=udf)
    >>> np.array(result["logsum"]).shape
    (32, 32)
    """
    def __init__(self):
        super().__init__()

    def get_backends(self):
        return [
            b for b in self.BACKEND_ALL
            if b not in (
                self.BACKEND_SCIPY_COO, self.BACKEND_SCIPY_CSR, self.BACKEND_SCIPY_CSC,
                self.BACKEND_CUPY_SCIPY_COO, self.BACKEND_CUPY_SCIPY_CSC,
                self.BACKEND_CUPY_SCIPY_CSR,
            )
        ]

    def get_result_buffers(self):
        ""
        return {
            'logsum': self.buffer(
                kind='sig', dtype='float32', where='device'
            ),
        }

    def merge(self, dest, src):
        ""
        dest.logsum[:] += src.logsum[:]

    def merge_all(self, ordered_results):
        ''
        chunks = [b.logsum for b in ordered_results.values()]
        logsum = np.stack(chunks, axis=0).sum(axis=0)
        return {'logsum': logsum}

    def process_frame(self, frame):
        ""
        self.results.logsum[:] += self.forbuf(
            np.log(frame - np.min(frame) + 1),
            self.results.logsum
        )


def run_logsum(ctx, dataset, roi=None):
    '''
    Sum up logscaled frames

    In comparison to log-scaling the sum, this highlights regions with slightly higher
    intensity that appear in many frames in relation to very high intensity in a few frames.

    Example:

    f1 = (11, 101)
    f2 = (11, 1)
    f2 = (11, 1)
    ...
    f10 = (11, 1)

    log10(sum(f1 ... f10)) == (2.04, 2.04)

    sum(log10(f1) ... log10(f10)) == (10.4, 2.04)

    '''
    udf = LogsumUDF()
    return ctx.run_udf(dataset=dataset, udf=udf, roi=roi)



import numpy as np
import pytest

from libertem.warnings import UseDiscouragedWarning
from libertem.common.exceptions import UDFException
from libertem.udf.base import UDF
from libertem.common.buffers import BufferWrapper


"""
Rules for buffer declarations, `get_result_buffers` and `get_results`:

1) All buffers are declared in `get_result_buffers`
2) If a buffer is only computed in `get_results`, it should be marked via
   `use='result_only'` so it isn't allocated on workers
3) If a buffer is only used as intermediary result, it should be marked via `use='private'`
4) Not including a buffer in `get_results` means it will either be passed on
   unchanged, or dropped if `use='private'`
5) It's an error to omit an `use='result_only'` buffer in `get_results`
6) It's an error to include a `use='private'` buffer in `get_results`
7) All results are returned from `Context.run_udf` as `BufferWrapper` instances
8) By default, if `get_results` is not implemented, `use='private'` buffers are dropped,
   and others are passed through unchanged
"""


class UndeclaredBufferUDF(UDF):
    def get_result_buffers(self):
        return {
            'buf': self.buffer(kind='nav', dtype=np.float32),
        }

    def process_frame(self, frame):
        pass

    def get_results(self):
        return {
            'buf': self.results.buf,
            'blah': np.zeros((128, 128), dtype=np.float64),
        }


def test_undeclared_buffer_error(lt_ctx, default_raw):
    """
    1) All buffers are declared in `get_result_buffers`
    """
    udf = UndeclaredBufferUDF()
    with pytest.raises(KeyError):
        lt_ctx.run_udf(dataset=default_raw, udf=udf)


class NoAllocateResultsUDF(UDF):
    def get_result_buffers(self):
        return {
            'buf1': self.buffer(kind='sig', dtype=np.float32),
            'buf2': self.buffer(kind='sig', dtype=np.float32, use='result_only'),
        }

    def merge(self, dest, src):
        assert 'buf2' not in dest
        assert 'buf2' not in src
        dest.buf1[:] += src.buf1

    def process_frame(self, frame):
        self.results.buf1[:] += frame
        assert self.results.buf2 is None

    def get_results(self):
        return {
            'buf1': self.results.buf1,
            'buf2': self.results.buf1 + 1,
        }


def test_no_allocate_result(lt_ctx, default_raw):
    """
    2) If a buffer is only computed in `get_results`, it should be marked via
       `use='result_only'` so it isn't allocated on workers
    """
    udf = NoAllocateResultsUDF()
    lt_ctx.run_udf(dataset=default_raw, udf=udf)


class UnifyResultTypesUDF(UDF):
    def get_result_buffers(self):
        return {
            'buf1': self.buffer(kind='sig', dtype=np.float32),
            'buf2': self.buffer(kind='sig', dtype=np.float32, use='result_only'),
        }

    def merge(self, dest, src):
        assert 'buf2' not in dest
        assert 'buf2' not in src
        dest.buf1[:] += src.buf1

    def process_frame(self, frame):
        pass

    def get_results(self):
        return {
            'buf1': self.results.buf1,
            'buf2': self.results.buf1 + 1,
        }


def test_unified_result_types(lt_ctx, default_raw):
    """
    7) All results are returned from `Context.run_udf` as `BufferWrapper` instances
    """
    udf = UnifyResultTypesUDF()
    results = lt_ctx.run_udf(dataset=default_raw, udf=udf)
    assert isinstance(results['buf1'], BufferWrapper)
    assert isinstance(results['buf2'], BufferWrapper)


class AverageUDF(UDF):
    """
    Like SumUDF, but also computes the average
    """
    def get_result_buffers(self):
        return {
            'sum': self.buffer(kind='sig', dtype=np.float32),
            'num_frames': self.buffer(kind='single', dtype=np.uint64),
            'average': self.buffer(kind='sig', dtype=np.float32, use='result_only'),
        }

    def process_frame(self, frame):
        self.results.sum[:] += frame
        self.results.num_frames[:] += 1

    def merge(self, dest, src):
        dest.sum[:] += src.sum
        dest.num_frames[:] += src.num_frames

    def get_results(self):
        avg = self.results.sum / self.results.num_frames
        return {
            'sum': self.results.sum,
            'average': avg,
        }


def test_delayed_buffer_alloc(lt_ctx, default_raw):
    udf = AverageUDF()
    results = lt_ctx.run_udf(dataset=default_raw, udf=udf)
    assert np.allclose(
        results['sum'].data / default_raw.shape.nav.size,
        results['average']
    )
    assert results['average'].dtype.kind == 'f'


def test_delayed_buffer_alloc_roi(lt_ctx, default_raw):
    udf = AverageUDF()
    roi = np.random.choice([True, False], size=default_raw.shape.nav)
    results = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    assert np.allclose(
        results['sum'].data / np.sum(roi),
        results['average']
    )


class SumSigUDFAndAHalf(UDF):
    def get_result_buffers(self):
        return {
            'sum': self.buffer(kind='nav', dtype=np.float32),
            'sum_half': self.buffer(kind='nav', dtype=np.float32, use='result_only'),
        }

    def process_frame(self, frame):
        self.results.sum[:] = np.sum(frame)

    def get_results(self):
        return {
            'sum': self.results.sum,
            'sum_half': self.results.sum / 2,
        }


def test_get_results_nav_with_roi(lt_ctx, default_raw):
    udf = SumSigUDFAndAHalf()
    roi = np.random.choice([True, False], size=default_raw.shape.nav)
    results = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    assert np.allclose(
        results['sum'].raw_data / 2,
        results['sum_half'].raw_data
    )
    assert np.allclose(
        results['sum'].data / 2,
        results['sum_half'].data,
        equal_nan=True,
    )


class UsePrivate1Defaults(UDF):
    def get_result_buffers(self):
        return {
            'default': self.buffer(kind='nav', dtype=np.float32),
            'private': self.buffer(kind='nav', dtype=np.float32, use='private'),
        }

    def process_frame(self, frame):
        self.results.default[:] = np.sum(frame)
        self.results.private[:] = np.sum(frame)


def test_use_private_defaults(lt_ctx, default_raw):
    """
    3) If a buffer is only used as intermediary result, it should be marked via `use='private'`
    8) By default, if `get_results` is not implemented, `use='private'` buffers are dropped,
       and others are passed through unchanged
    """
    udf = UsePrivate1Defaults()
    roi = np.random.choice([True, False], size=default_raw.shape.nav)
    results = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    assert 'default' in results
    assert 'private' not in results


class UsePrivate2ImplicitOmit(UsePrivate1Defaults):
    def get_results(self):
        return {
            'default': self.results.default,
        }


def test_use_private_implicit_omit(lt_ctx, default_raw):
    """
    3) If a buffer is only used as intermediary result, it should be marked via `use='private'`
    4) Not including a buffer in `get_results` means it will either be passed on
       unchanged, or dropped if `use='private'`
    """
    udf = UsePrivate2ImplicitOmit()
    roi = np.random.choice([True, False], size=default_raw.shape.nav)
    results = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    assert 'default' in results
    assert 'private' not in results


class UsePrivate3DontReturn(UsePrivate1Defaults):
    def get_results(self):
        return {
            'private': np.ones(self.meta.dataset_shape.nav),
        }


def test_use_private_dont_include_in_get_results(lt_ctx, default_raw):
    """
    6) It's an error to include a `use='private'` buffer in `get_results`
    """
    udf = UsePrivate3DontReturn()
    roi = np.random.choice([True, False], size=default_raw.shape.nav)
    with pytest.raises(UDFException):
        lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)


class BufferVisibilityDefault(UDF):
    def get_result_buffers(self):
        return {
            'default': self.buffer(kind='nav', dtype=np.float32),
            'private': self.buffer(kind='nav', dtype=np.float32, use='private'),
            'result_only': self.buffer(kind='nav', dtype=np.float32, use='result_only'),
        }

    def process_frame(self, frame):
        self.results.default[:] = np.sum(frame)
        self.results.private[:] = np.sum(frame)
        assert self.results.result_only is None

    # NOTE: no get_results implemented -> error because of the result_only buffer!


def test_buffer_visibility_default(lt_ctx, default_raw):
    """
    5) It's an error to omit an `use='result_only'` buffer in `get_results`
    """
    udf = BufferVisibilityDefault()
    roi = np.random.choice([True, False], size=default_raw.shape.nav)
    with pytest.raises(UDFException) as m:
        lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    assert m.match(
        "don't know how to set use='result_only'"
        " buffer 'result_only'; please implement `get_results`"
    )


class BufferVisibilityResOnly(BufferVisibilityDefault):
    def get_results(self):
        return {
            'result_only': self.results.default,  # just copy over the 'default' buffer
            # NOTE: other keys omitted, this should pass on the non-private keys
        }


def test_buffer_visibility_with_result_only(lt_ctx, default_raw):
    udf = BufferVisibilityResOnly()
    roi = np.random.choice([True, False], size=default_raw.shape.nav)
    results = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    assert 'result_only' in results
    assert 'default' in results, 'the use=None buffer should be implicitly included in the result'
    assert 'private' not in results

    for k in results:
        np.array(results[k])
        results[k].raw_data
        results[k].data


class ResultOnlyWithFullShape(UDF):
    def get_result_buffers(self):
        return {
            'default': self.buffer(kind='nav', dtype=np.float32),
            'result_only': self.buffer(kind='nav', dtype=np.float32, use='result_only'),
        }

    def process_frame(self, frame):
        self.results.default[:] = np.sum(frame)

    def get_results(self):
        # we want to embed the results ourselves... does this work?
        resonly = np.zeros(self.meta.dataset_shape.nav, dtype=self.results.default.dtype)
        resonly[self.meta.roi] = self.results.default
        return {
            'result_only': resonly,
        }


def test_get_results_nav_with_roi_full_shape(lt_ctx, default_raw):
    udf = ResultOnlyWithFullShape()
    roi = np.random.choice([True, False], size=default_raw.shape.nav)
    results = lt_ctx.run_udf(dataset=default_raw, udf=udf, roi=roi)
    assert 'result_only' in results
    assert np.allclose(
        results['default'].data[roi],
        results['result_only'].data[roi]
    )
    assert np.allclose(results['result_only'].data[~roi], 0)


class OldDictMergeAccess(UDF):
    def get_result_buffers(self):
        return {
            'default': self.buffer(kind='nav', dtype=np.float32),
        }

    def process_frame(self, frame):
        self.results.default[:] = np.sum(frame)

    def merge(self, dest, src):
        dest['default'][:] = src['default'][:]


def test_warning_for_dict_access_merge(lt_ctx, default_raw):
    udf = OldDictMergeAccess()
    with pytest.warns(UseDiscouragedWarning):
        lt_ctx.run_udf(dataset=default_raw, udf=udf)


class OldUDFDataDictAccess(UDF):
    def get_result_buffers(self):
        return {
            'default': self.buffer(kind='nav', dtype=np.float32),
        }

    def process_frame(self, frame):
        self.results.default[:] = np.sum(frame)

    def postprocess(self):
        self.results['default'].raw_data


def test_warning_for_dict_access_postprocess(lt_ctx, default_raw):
    udf = OldUDFDataDictAccess()
    with pytest.warns(UseDiscouragedWarning):
        lt_ctx.run_udf(dataset=default_raw, udf=udf)



from .sum import SumAnalysis
from .com import COMAnalysis
from .radialfourier import RadialFourierAnalysis
from .disk import DiskMaskAnalysis
from .ring import RingMaskAnalysis
from .point import PointMaskAnalysis
from .masks import MasksAnalysis
from .raw import PickFrameAnalysis
from .fem import FEMAnalysis
from .rawfft import PickFFTFrameAnalysis
from .sumfft import SumfftAnalysis
from .apply_fft_mask import ApplyFFTMask
from .sd import SDAnalysis
from .sumsig import SumSigAnalysis
from .clust import ClusterAnalysis

__all__ = [
    'SumAnalysis',
    'COMAnalysis',
    'RadialFourierAnalysis',
    'DiskMaskAnalysis',
    'RingMaskAnalysis',
    'PointMaskAnalysis',
    'MasksAnalysis',
    'PickFrameAnalysis',
    'FEMAnalysis',
    'PickFFTFrameAnalysis',
    'SumfftAnalysis',
    'ApplyFFTMask',
    'SDAnalysis',
    'ClusterAnalysis',
    'SumSigAnalysis'
]



from .code_template import CodeTemplate


def copy_notebook(conn, dataset, comp):
    instance = CodeTemplate(conn, dataset, comp)
    analy = []
    for _, analysis, plot, _ in instance.analysis():
        analy.append({
            'analysis': analysis,
            'plot': plot
        })

    return {
        'dependency': instance.dependency(),
        'initial_setup': instance.initial_setup(),
        'ctx': instance.connection()[0],
        'dataset': instance.dataset(),
        'analysis': analy
    }



from .base import BaseAnalysis, AnalysisResult, AnalysisResultSet
import libertem.udf.sumsigudf as sumsigudf
from .helper import GeneratorHelper


class SumSigTemplate(GeneratorHelper):

    short_name = "sumsig"

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return ["from libertem.analysis import SumSigAnalysis"]

    def get_docs(self):
        docs = ["# SumSig Analysis"]
        return '\n'.join(docs)

    def get_analysis(self):
        temp_analysis = [f"sumsig_analysis = SumSigAnalysis(dataset=ds, parameters={self.params})",
                         "sumsig_result = ctx.run(sumsig_analysis, progress=True)"]
        return '\n'.join(temp_analysis)

    def get_plot(self):
        plot = ["plt.figure()",
                "plt.imshow(sumsig_result.intensity.raw_data)",
                "plt.colorbar()"]
        return ['\n'.join(plot)]


class SumSigAnalysis(BaseAnalysis, id_="SUM_SIG"):
    TYPE = "UDF"

    def get_udf(self):
        return sumsigudf.SumSigUDF()

    def get_udf_results(self, udf_results, roi, damage):
        from libertem.viz import visualize_simple
        return AnalysisResultSet([
            AnalysisResult(raw_data=udf_results['intensity'],
                           visualized=visualize_simple(
                               udf_results['intensity'].data.reshape(self.dataset.shape.nav),
                               damage=damage
                           ),
                           key="intensity", title="intensity",
                           desc="result from frame integration"),
        ])

    @classmethod
    def get_template_helper(cls):
        return SumSigTemplate



import io
import logging
import inspect

import tornado.web
from libertem.analysis.base import Analysis
from libertem.web.engine import JobEngine

from libertem.web.rpc import RPCContext

from .base import CORSMixin, log_message
from .messages import Message
from .state import SharedState
from libertem.io.writers.results.base import ResultFormatRegistry

logger = logging.getLogger(__name__)


class AnalysisDetailHandler(CORSMixin, tornado.web.RequestHandler):
    """
    API Handler for CRUD of analyses.

    Each Analysis has a connection to a Dataset, and
    zero or one Job.

    """
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    async def put(self, compoundUuid, uuid):
        """
        Register or update an analysis, which is then later associated
        to their results.

        request body contains the analysis type and parameters

        On update, we let the currently running jobs continue running,
        jobs are handled separately by the JobDetailHandler.

        Update can change both parameters and analysis type.
        """
        request_data = tornado.escape.json_decode(self.request.body)
        dataset_id = request_data["dataset"]
        details = request_data['details']
        params = details['parameters']
        analysis_type = details['analysisType']
        existing_analysis = self.state.analysis_state.get(uuid)
        if existing_analysis is None:
            return await self._create(uuid, dataset_id, analysis_type, params)
        else:
            return await self._update(
                uuid, dataset_id, analysis_type, existing_analysis, params
            )

    async def _create(self, uuid, dataset_id, analysis_type, params):
        self.state.analysis_state.create(uuid, dataset_id, analysis_type, params)
        msg = Message().create_analysis(uuid, dataset_id, analysis_type, params)
        log_message(msg)
        self.event_registry.broadcast_event(msg)
        self.write(msg)

    async def _update(self, uuid, dataset_id, analysis_type, existing_analysis, params):
        self.state.analysis_state.update(uuid, analysis_type, params)
        msg = Message().update_analysis(uuid, dataset_id, analysis_type, params)
        log_message(msg)
        self.event_registry.broadcast_event(msg)
        self.write(msg)

    async def delete(self, compoundUuid, uuid):
        """
        Remove an analysis, stop all related jobs and remove
        analysis results.
        """
        result = await self.state.analysis_state.remove(uuid)
        if result:
            msg = Message().analysis_removed(uuid)
        else:
            # FIXME: concrete error message?
            msg = Message().analysis_removal_failed(uuid, "analysis could not be removed")
        log_message(msg)
        self.event_registry.broadcast_event(msg)
        self.write(msg)


class DownloadDetailHandler(CORSMixin, tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    def _get_format(self):
        # FIXME: unused for now
        fmt = self.request.arguments['fmt']
        assert len(fmt) == 1
        return fmt[0].decode("utf8")

    async def get(self, compoundUuid: str, uuid: str, file_format_id: str):
        details, result_set, job_id, udf_results = self.state.analysis_state.get_results(uuid)
        format_cls = ResultFormatRegistry.get_format_by_id(file_format_id)
        result_formatter = format_cls(result_set)
        buf = io.BytesIO()
        result_formatter.serialize_to_buffer(buf)

        self.set_header('Content-Type', result_formatter.get_content_type())
        self.set_header(
            'Content-Disposition',
            'attachment; filename="%s"' % result_formatter.get_filename(),
        )
        # FIXME: stream file (maybe temporary file w/ sendfile?)
        self.write(buf.getvalue())


class CompoundAnalysisHandler(CORSMixin, tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    async def put(self, uuid):
        request_data = tornado.escape.json_decode(self.request.body)
        dataset_id = request_data['dataset']
        details = request_data['details']
        main_type = details['mainType']
        analyses = details['analyses']
        created = self.state.compound_analysis_state.create_or_update(
            uuid, main_type, dataset_id, analyses
        )
        serialized = self.state.compound_analysis_state.serialize(uuid)
        if created:
            msg = Message().compound_analysis_created(serialized)
        else:
            msg = Message().compound_analysis_updated(serialized)
        log_message(msg)
        self.event_registry.broadcast_event(msg)
        self.write(msg)

    async def delete(self, uuid):
        ca = self.state.compound_analysis_state[uuid]
        for analysis_id in ca["details"]["analyses"]:
            result = await self.state.analysis_state.remove(analysis_id)
            if result:
                msg = Message().analysis_removed(analysis_id)
            else:
                # FIXME: concrete error message?
                msg = Message().analysis_removal_failed(
                    analysis_id, "analysis could not be removed"
                )
            log_message(msg)
            self.event_registry.broadcast_event(msg)

        self.state.compound_analysis_state.remove(uuid)

        msg = Message().compound_analysis_removed(uuid)
        log_message(msg)
        self.event_registry.broadcast_event(msg)
        self.write(msg)


class AnalysisRPCHandler(CORSMixin, tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry
        self.engine = JobEngine(state, event_registry)

    async def put(self, compound_analysis_id: str, proc_name: str):
        rpc_context = RPCContext(
            state=self.state,
            compound_analysis_id=compound_analysis_id,
            engine=self.engine,
        )
        comp_ana = rpc_context.get_compound_analysis()
        ana_type = comp_ana['details']['mainType']
        analysis_cls = Analysis.get_analysis_by_type(ana_type)
        rpc_def = analysis_cls.get_rpc_definitions()
        if proc_name not in rpc_def:
            self.set_status(400, "Bad request: unknown RPC method")
            self.write({
                "status": "error",
                "msg": "unknown RPC method",
            })
            return

        ProcCls = rpc_def[proc_name]

        proc = ProcCls()
        # support both sync and async variants here:
        if inspect.iscoroutinefunction(proc.__call__):
            result = await proc(rpc_context)
        else:
            return proc(rpc_context)
        self.write(result)



import io
import os
import sys
import typing
from typing import Optional
import logging

import numpy as np
from numpy.lib.utils import safe_eval
from numpy.lib.format import read_magic
from numpy.compat import long, asstr
from libertem.common.messageconverter import MessageConverter

from libertem.io.dataset.base import (
    DataSet, FileSet, BasePartition, DataSetException, DataSetMeta, File, IOBackend,
)
from libertem.common import Shape
from libertem.common.math import prod

log = logging.getLogger(__name__)


class NPYDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/NPYDatasetParams.schema.json",
        "title": "NPYDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "NPY"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
            "io_backend": {
                "enum": IOBackend.get_supported(),
            },
        },
        "required": ["type", "path"],
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = int(raw_data["sync_offset"])
        return data


class NPYInfo(typing.NamedTuple):
    dtype: str
    shape: tuple[int, ...]
    count: int
    offset: int


# `_read_bytes`, `_read_array_header`, `_filter_header` are
# stolen from `numpy.lib.format`, as they don't appear to be
# part of the public API.

def _read_bytes(fp, size, error_template="ran out of data"):
    """
    Read from file-like object until size bytes are read.
    Raises ValueError if not EOF is encountered before size bytes are read.
    Non-blocking objects only supported if they derive from io objects.

    Required as e.g. ZipExtFile in python 2.6 can return less data than
    requested.
    """
    data = b''
    while True:
        # io files (default in python3) return None or raise on
        # would-block, python2 file will truncate, probably nothing can be
        # done about that.  note that regular files can't be non-blocking
        try:
            r = fp.read(size - len(data))
            data += r
            if len(r) == 0 or len(data) == size:
                break
        except io.BlockingIOError:
            pass
    if len(data) != size:
        msg = "EOF: reading %s, expected %d bytes got %d"
        raise ValueError(msg % (error_template, size, len(data)))
    else:
        return data


def _read_array_header(fp, version):
    """
    see read_array_header_1_0
    """
    # Read an unsigned, little-endian short int which has the length of the
    # header.
    import struct
    if version == (1, 0):
        hlength_str = _read_bytes(fp, 2, "array header length")
        header_length = struct.unpack('<H', hlength_str)[0]
        header = _read_bytes(fp, header_length, "array header")
    elif version == (2, 0):
        hlength_str = _read_bytes(fp, 4, "array header length")
        header_length = struct.unpack('<I', hlength_str)[0]
        header = _read_bytes(fp, header_length, "array header")
    else:
        raise ValueError("Invalid version %r" % version)

    # The header is a pretty-printed string representation of a literal
    # Python dictionary with trailing newlines padded to a 16-byte
    # boundary. The keys are strings.
    #   "shape" : tuple of int
    #   "fortran_order" : bool
    #   "descr" : dtype.descr
    header = _filter_header(header)
    try:
        d = safe_eval(header)
    except SyntaxError as e:
        msg = "Cannot parse header: %r\nException: %r"
        raise ValueError(msg % (header, e))
    if not isinstance(d, dict):
        msg = "Header is not a dictionary: %r"
        raise ValueError(msg % d)
    keys = sorted(d.keys())
    if keys != ['descr', 'fortran_order', 'shape']:
        msg = "Header does not contain the correct keys: %r"
        raise ValueError(msg % (keys,))

    # Sanity-check the values.
    if (not isinstance(d['shape'], tuple)
            or not np.all([isinstance(x, (int, long)) for x in d['shape']])):
        msg = "shape is not valid: %r"
        raise ValueError(msg % (d['shape'],))
    if not isinstance(d['fortran_order'], bool):
        msg = "fortran_order is not a valid bool: %r"
        raise ValueError(msg % (d['fortran_order'],))
    try:
        dtype = np.dtype(d['descr'])
    except TypeError:
        msg = "descr is not a valid dtype descriptor: %r"
        raise ValueError(msg % (d['descr'],))

    return d['shape'], d['fortran_order'], dtype


def _filter_header(s):
    """Clean up 'L' in npz header ints.

    Cleans up the 'L' in strings representing integers. Needed to allow npz
    headers produced in Python2 to be read in Python3.

    Parameters
    ----------
    s : byte string
        Npy file header.

    Returns
    -------
    header : str
        Cleaned up header.

    """
    import tokenize
    if sys.version_info[0] >= 3:
        from io import StringIO
    else:
        from StringIO import StringIO

    tokens = []
    last_token_was_number = False
    for token in tokenize.generate_tokens(StringIO(asstr(s)).read):
        token_type = token[0]
        token_string = token[1]
        if (last_token_was_number
                and token_type == tokenize.NAME
                and token_string == "L"):
            continue
        else:
            tokens.append(token)
        last_token_was_number = (token_type == tokenize.NUMBER)
    return tokenize.untokenize(tokens)


def read_npy_info(path: str) -> NPYInfo:
    with open(path, "rb") as fp:
        version = read_magic(fp)
        shape, fortran_order, dtype = _read_array_header(fp, version)
        if fortran_order:
            raise DataSetException('Unable to process Fortran-ordered NPY arrays, '
                                   'consider converting with np.ascontiguousarray().')
        if len(shape) == 0:
            count = 1
        else:
            count = int(np.multiply.reduce(shape, dtype=np.int64))
        offset = fp.tell()
        return NPYInfo(dtype=dtype, shape=shape, count=count, offset=offset)


class NPYDataSet(DataSet):
    """
    .. versionadded:: 0.10.0

    Read data stored in a NumPy .npy binary file. Dataset shape
    and dtype are inferred from the file header unless overridden
    by the arguments to this class.

    As of this time Fortran-ordered .npy files are not supported

    Parameters
    ----------
    path : str
        The path to the .npy file
    sig_dims : int, optional, by default 2
        The number of dimensions from the end of the full shape
        to interpret as signal dimensions. If None
        will be inferred from the sig_shape argument when present.
    nav_shape : Tuple[int, int], optional
        A nav_shape to apply to the dataset overriding the shape
        value read from the .npy header, by default None. This can
        be used to read a subset of the .npy file, or reshape the
        contained data. Frames are read in C-order from the beginning
        of the file.
    sig_shape : Tuple[int, int], optional
        A sig_shape to apply to the dataset overriding the shape
        value read from the .npy header, by default None. Pixels are
        read in C-order from the beginning of the file.
    sync_offset : int, optional, by default 0
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start
    io_backend : IOBackend, optional
        The I/O backend to use, see :ref:`io backends`, by default None.

    Raises
    ------
    DataSetException
        If sig_dims is not an integer and cannot be inferred from sig_shape
    DataSetException
        If the supplied nav_shape + sig_shape describe an array larger
        than the contents of the .npy file
    DataSetException
        If the .npy file is Fortran-ordered

    Examples
    --------

    >>> ds = ctx.load("npy", path='./path_to_file.npy')  # doctest: +SKIP
    """
    def __init__(
        self,
        path: str,
        sig_dims: Optional[int] = 2,
        nav_shape: Optional[tuple[int, int]] = None,
        sig_shape: Optional[tuple[int, int]] = None,
        sync_offset: int = 0,
        io_backend: Optional[IOBackend] = None,
    ):
        super().__init__(io_backend=io_backend)
        self._meta = None
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sig_dims = sig_dims
        if self._sig_shape is not None:
            if self._sig_dims is None:
                self._sig_dims = len(self._sig_shape)
            if len(self._sig_shape) != self._sig_dims:
                raise DataSetException(f'Mismatching sig_dims (= {self._sig_dims}) '
                                       f'and sig_shape {self._sig_shape} arguments')
        if self._sig_dims is None or not isinstance(self._sig_dims, int):
            raise DataSetException('Must supply one of sig_dims or sig_shape to NPYDataSet')
        self._path = path
        self._sync_offset = sync_offset
        self._npy_info: typing.Optional[NPYInfo] = None

    def _get_filesize(self):
        return os.stat(self._path).st_size

    def initialize(self, executor) -> "DataSet":
        self._filesize = executor.run_function(self._get_filesize)
        npyinfo = executor.run_function(read_npy_info, self._path)
        self._npy_info = npyinfo
        np_shape = Shape(npyinfo.shape, sig_dims=self._sig_dims)
        sig_shape = self._sig_shape if self._sig_shape else np_shape.sig
        nav_shape = self._nav_shape if self._nav_shape else np_shape.nav
        shape = Shape(tuple(nav_shape) + tuple(sig_shape), sig_dims=self._sig_dims)
        # Trying to follow implementation of RawFileDataSet i.e. the _image_count
        # is the whole block of data interpreted as N frames of sig_shape, noting that
        # here sig_shape can be either user-supplied or from the npy metadata
        # if prod(sig_shape) is not a factor then bytes will be dropped at the end
        self._image_count = np_shape.size // prod(sig_shape)
        self._nav_shape_product = shape.nav.size
        self._sync_offset_info = self.get_sync_offset_info()
        self._meta = DataSetMeta(
            shape=shape,
            raw_dtype=np.dtype(npyinfo.dtype),
            sync_offset=self._sync_offset or 0,
            image_count=self._image_count,
        )
        return self

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    @classmethod
    def detect_params(cls, path, executor):
        try:
            npy_info = executor.run_function(read_npy_info, path)
            # FIXME: assumption about number of sig dims
            shape = Shape(npy_info.shape, sig_dims=2)
            return {
                "parameters": {
                    "path": path,
                    "nav_shape": tuple(shape.nav),
                    "sig_shape": tuple(shape.sig),
                },
                "info": {
                    "image_count": int(prod(shape.nav)),
                    "native_sig_shape": tuple(shape.sig),
                }
            }
        except Exception as e:
            print(e)
            return False

    def get_diagnostics(self):
        return [
            {"name": "dtype", "value": str(self.dtype)}
        ]

    @classmethod
    def get_supported_extensions(cls):
        return {"npy"}

    @classmethod
    def get_msg_converter(cls):
        return NPYDatasetParams

    def _get_fileset(self):
        assert self._npy_info is not None
        return FileSet([
            File(
                path=self._path,
                start_idx=0,
                end_idx=self._meta.image_count,
                sig_shape=self.shape.sig,
                native_dtype=self._meta.raw_dtype,
                file_header=self._npy_info.offset,
            )
        ])

    def check_valid(self):
        try:
            fileset = self._get_fileset()
            backend = self.get_io_backend().get_impl()
            with backend.open_files(fileset):
                return True
        except (OSError, ValueError) as e:
            raise DataSetException("invalid dataset: %s" % e)

    def get_cache_key(self):
        return {
            "path": self._path,
            # nav_shape + sig_shape; included because changing nav_shape will change
            # the partition structure and cause errors
            "shape": tuple(self.shape),
            "dtype": str(self.dtype),
            "sync_offset": self._sync_offset,
        }

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield BasePartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )

    def __repr__(self):
        return f"<NPYDataSet of {self.dtype} shape={self.shape}>"



from typing import Optional

from sparseconverter import ArrayBackend, NUMPY, for_backend
import dask
import dask.array

from libertem.io.dataset.base import DataSet, DataTile


def _get_tile_straight(p, dest_dtype, roi, array_backend):
    return p.get_macrotile(dest_dtype=dest_dtype, roi=roi, array_backend=array_backend)


def _get_tile_converted(p, dest_dtype, roi, array_backend):
    arr_tile = p.get_macrotile(dest_dtype=dest_dtype, roi=roi, array_backend=None)
    converted_chunk = for_backend(arr_tile.data, array_backend)
    return DataTile(
        data=converted_chunk.reshape(arr_tile.shape),
        tile_slice=arr_tile.tile_slice,
        scheme_idx=arr_tile.scheme_idx
    )


def make_dask_array(dataset: DataSet, dtype='float32', roi=None,
        array_backend: Optional[ArrayBackend] = None):
    '''
    Create a Dask array using the DataSet's partitions as blocks.
    '''
    chunks = []
    workers = {}
    if array_backend is None:
        array_backend = NUMPY
    if array_backend in dataset.array_backends:
        get_tile = _get_tile_straight
    else:
        get_tile = _get_tile_converted

    for p in dataset.get_partitions():
        d = dask.delayed(get_tile)(
            p=p, dest_dtype=dtype, roi=roi, array_backend=array_backend
        ).data
        workers[d] = p.get_locations()
        chunks.append(
            dask.array.from_delayed(
                d,
                dtype=dtype,
                shape=p.slice.adjust_for_roi(roi).shape,
            )
        )
    arr = dask.array.concatenate(chunks, axis=0)
    if roi is None:
        arr = arr.reshape(dataset.shape)
    return (arr, workers)



from io import BytesIO
from typing import (
    Callable, Optional, Union
)

import numpy as np


class AnalysisResult:
    """
    This class represents a single 2D image result from an Analysis.

    Instances of this class are contained in an :class:`AnalysisResultSet`.

    Attributes
    ----------
    raw_data : numpy.ndarray
        The raw numerical data of this result
    visualized : numpy.ndarray
        Visualized result as :class:`numpy.ndarray` with RGB or RGBA values
    title : str
        Title for the GUI
    desc : str
        Short description in the GUI
    key : str
        Key to identify the result in an :class:`AnalysisResultSet`
    """
    def __init__(
        self,
        raw_data: np.ndarray,
        visualized: np.ndarray,
        title: str,
        desc: str,
        key: str,
        include_in_download: bool = True,
    ):
        self.include_in_download = include_in_download
        self.raw_data = raw_data
        self._visualized = visualized
        self.title = title
        self.desc = desc
        self.key = key

    def __str__(self):
        result = ""
        for k in ("title", "desc", "key", "raw_data", "visualized"):
            result += f"{k}: {getattr(self, k)}\n"
        return result

    def __repr__(self):
        return "<AnalysisResult: %s>" % self.key

    def __array__(self):
        return np.array(self.raw_data)

    def get_image(self, save_kwargs: Optional[dict] = None) -> BytesIO:
        from libertem.common.viz import encode_image
        return encode_image(self.visualized, save_kwargs=save_kwargs)

    @property
    def visualized(self):
        if callable(self._visualized):
            self._visualized = self._visualized()
        return self._visualized


_ResultsType = Union[list[AnalysisResult], Callable[[], list[AnalysisResult]]]


class AnalysisResultSet:
    """
    Base class for Analysis result sets. :meth:`libertem.api.Context.run`
    returns an instance of this class or a subclass. Many of the subclasses are
    just introduced to document the Analysis-specific results (keys) of the
    result set and don't introduce new functionality.

    The contents of an :class:`AnalysisResultSet` can be addressed in four different ways:

    1. As a container class with the :attr:`AnalysisResult.key` properties as attributes
       of type :class:`AnalysisResult`.
    2. As a list of :class:`AnalysisResult` objects.
    3. As an iterator of :class:`AnalysisResult` objects (since 0.3).
    4. As a dictionary of :class:`AnalysisResult` objects with the :attr:`AnalysisResult.key`
       properties as keys.

    This allows to implement generic result handling code for an Analysis, for example GUI display,
    as well as specific code for particular Analysis subclasses.

    Attributes
    ----------
    raw_results
        Raw results from the underlying numerical processing, if available, otherwise None.

    Examples
    --------
    >>> mask_shape = tuple(dataset.shape.sig)
    >>> def m0():
    ...    return np.ones(shape=mask_shape)
    >>> def m1():
    ...     result = np.zeros(shape=mask_shape)
    ...     result[0,0] = 1
    ...     return result
    >>> analysis = ctx.create_mask_analysis(
    ...     dataset=dataset, factories=[m0, m1]
    ... )
    >>> result = ctx.run(analysis)
    >>> # As an object with attributes
    >>> print(result.mask_0.title)
    mask 0
    >>> print(result.mask_1.title)
    mask 1
    >>> # As a list
    >>> print(result[0].title)
    mask 0
    >>> # As an iterator
    >>> # New since 0.3.0
    >>> for m in result:
    ...     print(m.title)
    mask 0
    mask 1
    >>> # As a dictionary
    >>> print(result['mask_1'].title)
    mask 1
    """
    def __init__(self, results: _ResultsType, raw_results=None):
        self._results = results
        self.raw_results = raw_results

    def __repr__(self):
        return repr(self.results)

    def __getattr__(self, k):
        for result in self.results:
            if result.key == k:
                return result
        raise AttributeError("result with key '{}' not found, have: {}".format(
            k, ", ".join([r.key for r in self.results])
        ))

    def __getitem__(self, k):
        if isinstance(k, str):
            return self.__getattr__(k)
        return self.results[k]

    def __len__(self):
        return len(self.results)

    def keys(self):
        return [r.key for r in self.results]

    @property
    def results(self):
        if callable(self._results):
            self._results = self._results()
        return self._results

    def __iter__(self):
        return iter(self.results)



# keep version and revision in separate files such that we can
# generate them without any regex magic

from .__version__ import __version__

try:
    from ._baked_revision import revision
except ImportError:
    from .versioning import get_git_rev
    revision = get_git_rev()

__all__ = [
    "revision", "__version__"
]



import numpy as np

from libertem.common import Shape
from libertem.udf.base import UDF
from libertem.common.math import prod
from libertem.common.buffers import reshaped_view


class RecordUDF(UDF):
    '''
    Record input data as NumPy .npy file

    Parameters
    ----------

    filename : str or path-like
        Filename where to save. The file will be overwritten if it exists.
    _is_master : bool
        Internal flag, keep at default value.
    '''
    def __init__(self, filename, _is_master=True):
        self._is_master = _is_master
        super().__init__(filename=filename, _is_master=False)

    def get_preferred_input_dtype(self):
        ''
        return self.USE_NATIVE_DTYPE

    @property
    def _ds_shape(self) -> Shape:
        # only valid during run_udf hence _ property
        return self.meta.dataset_shape

    @property
    def _memmap_flat_shape(self) -> tuple[int, ...]:
        # only valid during run_udf hence _ property
        return (prod(self._ds_shape.nav), *self._ds_shape.sig)

    def preprocess(self):
        ''
        if self.meta.roi is not None:
            raise RuntimeError('Recording with ROI is not supported.')
        # create the file once in the preprocess method on the master node
        if self._is_master:
            np.lib.format.open_memmap(
                self.params.filename,
                mode='w+',
                dtype=self.meta.input_dtype,
                shape=tuple(self._ds_shape),
            )

    def get_result_buffers(self):
        ''
        return {}

    def get_task_data(self):
        ''
        m = np.lib.format.open_memmap(
                self.params.filename,
                mode='r+',
                dtype=self.meta.input_dtype,
                shape=tuple(self._ds_shape)
        )
        return {
            'memmap': reshaped_view(m, self._memmap_flat_shape)
        }

    def process_tile(self, tile):
        ''
        self.meta.slice.get(self.task_data.memmap)[:] = tile



import pytest
import numpy as np

from libertem.analysis.base import AnalysisResultSet, AnalysisResult


def test_result_set():
    result = AnalysisResult(
        raw_data=None,
        visualized=None,
        title="test",
        desc="test",
        key="test"
    )

    results = AnalysisResultSet([result])

    with pytest.raises(AttributeError) as einfo:
        results.foo
    assert einfo.match("not found")
    assert einfo.match("have: test")

    assert results.test == result
    assert len(results) == 1
    assert results[0] == result
    for item in results:
        assert item == result


def test_result_coerce_to_array():
    result = AnalysisResult(
        raw_data=(np.zeros((16, 16)), np.ones((16, 16))),
        visualized=None,
        title="test",
        desc="test",
        key="test"
    )

    # "object __array__ method not producing an array"
    assert np.array(result).shape == (2, 16, 16)
